import { RouteLocationNormalized, RouteRecord, RouteLocation, NavigationGuard, ListenerRemover, PostNavigationGuard } from './types';
import { RouterHistory } from './history/common';
import { ScrollToPosition, ScrollPosition } from './utils/scroll';
declare type ErrorHandler = (error: any) => any;
interface ScrollBehavior {
    (to: RouteLocationNormalized, from: RouteLocationNormalized, savedPosition: ScrollToPosition | null): ScrollPosition | Promise<ScrollPosition>;
}
export interface RouterOptions {
    history: RouterHistory;
    routes: RouteRecord[];
    scrollBehavior?: ScrollBehavior;
}
export interface Router {
    currentRoute: Readonly<RouteLocationNormalized>;
    resolve(to: RouteLocation): RouteLocationNormalized;
    createHref(to: RouteLocationNormalized): string;
    push(to: RouteLocation): Promise<RouteLocationNormalized>;
    replace(to: RouteLocation): Promise<RouteLocationNormalized>;
    doInitialNavigation(): Promise<void>;
    beforeEach(guard: NavigationGuard): ListenerRemover;
    afterEach(guard: PostNavigationGuard): ListenerRemover;
    onError(handler: ErrorHandler): void;
    onReady(): Promise<void>;
}
export declare function createRouter({ history, routes, scrollBehavior, }: RouterOptions): Router;
export {};
